### PersefoneDesktop - a GUI to the Persefone model of agriculture and ecosystems
###
### This file links the QML UI designs to the underlying model.
###


function loadsimulation(filename)
    global model, landcovermap
    @emit showSplash()
    fn = convert(filename, String)
    startswith(fn, "file://") && (fn = fn[8:end])
    running[] = false
    model = loadmodelobject(fn)
    landcovermap = rotr90(load(@param(world.landcovermap)))
    date[] = model.date
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit closeSplash()
    @emit updateMakie()
    println("Loaded simulation.")
end

function savesimulation(filename)
    global model
    fn = convert(filename, String)
    startswith(fn, "file://") && (fn = fn[8:end])
    savemodelobject(model, fn)
    println("Saved simulation.")
end

function Base.convert(s::QString, ::Type{<:AbstractString})
    s2 = ""
    for l in s
        s2 *= l
    end
    s2
end

function saveoutput()
    global model
    println("Saving output to file")
    if !@param(core.csvoutput) #override preference
        for output in model.dataoutputs
            CSV.write(joinpath(@param(core.outdir), output.name*".csv"),
                      model.datatables[output.name])
        end
    end
    visualiseoutput(model)
    GLMakie.activate!() # visualiseoutput() switches to Cairo
end

# I'm reimplementing `visualisemap()` and populationtrends() here, because I need
# to use Observables to keep the figure updated (if I use multiple plots, I get a
# memory leak, see here: https://github.com/MakieOrg/Makie.jl/issues/795)

function render_map(screen)
    global model, mapimage
    launching[] && return display(screen, Figure().scene) # blank screen at launch
    println("Updating map")
    if isnothing(mapimage)
        landcover = rotr90(load(joinpath(@param(world.mapdirectory), @param(world.landcovermap))))
        mapimage = Figure()
        ax = Axis(mapimage[1,1])
        hidedecorations!(ax)
        image!(mapimage[1,1], landcover)
        ax.aspect = DataAspect()
        inds = @lift(@transform!(@subset(@data("individuals"),
                                         :X .!= -1, :Date .== $date-Day(1)),
                                 :Y = size(landcover)[2] .- :Y))
        ncolors = max(2, length(@param(nature.targetspecies)))
        update_theme!(palette=(color=cgrad(:seaborn_bright, ncolors),),
                      cycle=[:color])
        for s in @param(nature.targetspecies)
            points = @lift(Matrix{Float32}(
                @select!(@subset($inds, :Species .== s), :X, :Y)))
            scatter!(mapimage[1,1], points, markersize=8)
        end
    end
    display(screen, mapimage.scene)
end
   
function render_plot(screen)
    global model, plotimage
    launching[] && return display(screen, Figure().scene) # blank screen at launch
    println("Updating plot")
    if isnothing(plotimage)
        pops = @lift begin
            $date #needed to activate the lift, doesn't do anything
            @data("populations")
        end
        #XXX other colour schemes: :tab10, :Accent_8, :Dark2_8, :Paired_12, :Set1_9
        # https://juliagraphics.github.io/ColorSchemes.jl/stable/catalogue/
        ncolors = max(2, length(@param(nature.targetspecies)))
        update_theme!(palette=(color=cgrad(:seaborn_bright, ncolors),), cycle=[:color])
        plotimage = Figure()
        dates = @param(core.startdate):@param(core.enddate)
        axlimits = @lift((1, length(dates), 0, maximum($pops[!,:Abundance])))
        ax = Axis(plotimage[1,1], xlabel="Date", ylabel="Population size",
                  limits=axlimits, xticks=Persefone.gettickmarks(dates))
        for s in @param(nature.targetspecies)
            points = @lift(Vector{Float32}(@subset($pops, :Species .== s).Abundance))
            iszero(size(points[])[1]) && continue
            lines!(plotimage[1,1], points, linewidth=3, label=s)
        end
        #FIXME the legend does something weird when run live
        size(pops[])[1] > 0 && axislegend("Species"; position=:lt)
    end
    display(screen, plotimage.scene)
end

on(delay) do d
    println("Delay is now $(round(d, digits=1)) seconds.") #XXX DEBUG
end

on(launching) do l
    if l
        @emit showSplash()
        println("Showing splash screen.") #XXX DEBUG
    else
        @emit closeSplash()
        @emit updateMakie()
    end
end

on(running) do r
    if r
        runbuttontext[] = "||"
        runbuttontip[] = "Pause"
    else
        runbuttontext[] = ">>"
        runbuttontip[] = "Run"
    end
end

datestring = () -> Dates.format(date[], "dd U yyyy")

function activateqmlfunctions()
    #@qmlfunction configwindow
    @qmlfunction loadsimulation
    @qmlfunction savesimulation
    @qmlfunction nextstep
    @qmlfunction previousstep
    @qmlfunction datestring
    #@qmlfunction writeconfig
    @qmlfunction saveoutput
end

"""
    launch()

The main function that creates the application.
"""
function launch()
    global model, timer

    activateqmlfunctions()
    timer = QTimer()

    GLMakie.activate!()
    ENV["QSG_RENDER_LOOP"] = "basic"
    QML.setGraphicsApi(QML.OpenGL)
    
    qmlfile = joinpath(dirname(@__FILE__), "main.qml")
    loadqml(qmlfile,
            timer = timer,
            vars = JuliaPropertyMap(guiproperties...),
            render_map_callback = @safe_cfunction(render_map, Cvoid, (Any,)),
            render_plot_callback = @safe_cfunction(render_plot, Cvoid, (Any,)))
    isnothing(model) && (launching[] = true) # initiates model initialisation
    println("Launched Persefone Desktop.")
    exec()
end
